import json
import numpy as np
import matplotlib.pyplot as plt
from sklearn.cluster import DBSCAN
from typing import List, Dict, Any, TypeVar
from deprecated import deprecated

T = TypeVar('T')

class Clusterer:
    '''
    Clusterer for applying density-based clustering on datasets.
    The clustering is done with DBSCAN.

    :param epsilon: Epsilon used in DBSCAN 
    :param min_points: Min_points used in DBSCAN
    '''
    def __init__(self, epsilon=11, min_points=2):
        self.epsilon = epsilon
        self.min_points = min_points

    def draw_locations(self, locations:List, labels:List=None) -> plt.Figure:
        if locations is None or len(locations) == 0:
            return self._draw_locations()

        if labels is None or len(locations) != len(labels):
            labels = self.create_labels(locations)

        return self._draw_locations(
            locations = self.extract_location_data(locations),
            partition_info = labels
        )

    def _draw_locations(self, locations:np.ndarray=None, centroids:np.ndarray=None, partition_info:List=None) -> plt.Figure:
        fig = plt.Figure()
        axis = fig.add_subplot(1, 1, 1)

        if locations is not None:
            colors = plt.cm.rainbow(np.linspace(0, 1, len(locations)))

            if partition_info is not None:
                distinct_colors = plt.cm.rainbow(np.linspace(0, 1, len(set(partition_info))))
                colors = [distinct_colors[pi] for pi in partition_info]

            # draw locations with random colors
            axis.scatter(locations[:,0],
                        locations[:,1], 
                        c=colors)

        if centroids is not None:
            # draw black centroids
            axis.scatter(centroids[:,0], centroids[:,1], c='k', marker='x', s=80)
            
        return fig
    
    def create_labels(self, features:np.ndarray) -> List[int]:
        '''Creates labels for the items based on DBSCAN.'''
        if features is None or len(features) == 0:
            return features # trash in trash out
    
        dbsc = DBSCAN(eps = self.epsilon, min_samples = self.min_points)
        dbsc = dbsc.fit(features)
        labels = dbsc.labels_

        return labels.tolist()

    @deprecated(reason="Use generic version instead")
    def extract_location_features(self, locations: List[dict]) -> np.ndarray:
        return np.asarray([(float(l['latitude']), float(l['longitude'])) for l in locations])

    @deprecated(reason="Use generic version instead")
    def extract_time_features(self, times: List[Dict]) -> np.ndarray:
        return np.asarray([[float(t['timestamp'])] for t in times])

    def _extract_features(self, dataset: List[Dict], features:List[str]) -> np.ndarray:
        '''Extracts the feature values from the dataset into a np array with same order as original dataset.'''
        extracted_features = []
        for data in dataset:
            entry = [float(data[feature]) for feature in features]
            extracted_features.append(entry)

        return np.asarray(extracted_features)

    def label_dataset(self, dataset:List[Dict], labels:List[Any]) -> List:
        '''Adds the labels to the elements of the dataset at the same position. The new key is called cluster_label.'''
        if dataset is None or labels is None:
            return

        if len(dataset) != len(labels):
            raise ValueError("dataset and labels has to have same length")

        for i in range(len(dataset)):
            if 'cluster_label' in dataset[i]:
                continue
            dataset[i]['cluster_label'] = labels[i]

    def group_by_clusters(self, dataset:List[Dict], labels:List[T]) -> Dict[T, List[Dict]]:
        self.label_dataset(dataset, labels)

        clusters = {}
        for label in labels:
            clusters[label] = [ds for ds in dataset if ds['cluster_label'] == label]
        
        return clusters

    @deprecated(reason="Use generic version instead")
    def cluster_locations(self, locations:List[Dict]) -> Dict[int, List[Dict]]:
        '''Returns a dictionary with identified clusters and their locations copied from the input'''
        if locations is None or len(locations) == 0:
            # raise Exception("locations has to contain something")
            return {}

        features = self.extract_location_features(locations)
        
        labels = self.create_labels(features)
        self.label_dataset(locations, labels)

        return self.group_by_clusters(locations, labels)
       
    @deprecated(reason="Use generic version instead") 
    def cluster_times(self, times:List[Dict]) -> Dict[int, List[Dict]]:
        '''Returns a dictionary with identified clusters and their times copied from the input'''
        features = self.extract_time_features(times)
    
        labels = self.create_labels(features)
        self.label_dataset(times, labels)

        return self.group_by_clusters(times, labels)

    def cluster_dataset(self, dataset:List[Dict], features:List[str]) -> List:
        '''
        Returns the identified clusters containing a subset of nodes from the dataset.
        
        :param dataset: The nodes to assign to clusters
        :param features: The feature names of the nodes to use for clustering

        :returns: A list of clusters
        '''
        arr = self._extract_features(dataset, features)

        labels = self.create_labels(arr)

        return self.group_by_clusters(dataset, labels)


# TODO remove
if __name__ == '__main__':
    import sys
    sys.path.insert(1, './')
    from db.agi.agi_repository import AgiRepository

    clusterer = Clusterer()
    agi_repo = AgiRepository()

    if True:
        res_old = clusterer.cluster_locations(agi_repo.getLocationsBasedOnNewDataSchema()['Nodes'])
        # print(res_old[11])
        # [{'id': 'adad64cb-bd71-4b2b-9a70-e08eb8b19901-1570900602', 'latitude': -20.2695062, 'longitude': 57.6297389, 'timestamp': 1570900602, 'user': 'b57ad1fb396cfc18b8867fb2e08be723c2cdc2a6', 'cluster_label': 11}, {'id': '127af17b-e823-4d30-8227-00f5421bd48b-1549291309', 'latitude': -20.5362627, 'longitude': 47.2459749, 'timestamp': 1549291309, 'user': 'ca34bd51c4dc65cbc021cb27bcaa014ca082b8c4', 'cluster_label': 11}]

    data = agi_repo.getLocationsBasedOnNewDataSchema()
    res = clusterer.cluster_dataset(data['Nodes'], data['Properties'])
    # if res is not None:
    #     print(res[11])

    assert (res_old == res)

    # time 
    res_old = clusterer.cluster_times(agi_repo.getTimesBasedOnNewDataSchema()['Nodes'])

    data = agi_repo.getTimesBasedOnNewDataSchema()
    res = clusterer.cluster_dataset(data['Nodes'], data['Properties'])
    print(res_old[20])
    print(res[20])
    
    assert (res_old == res)    